/*
 *				**** EVENT ****
 *
 * This program is an example of a simple event driven
 * program.  A real event driven program would be much
 * more complex, and might even have several programs
 * running at the same time and 'talking' to each other
 *
 * THIS PROGRAM WILL ONLY RUN UNDER MS-DOS
 * It uses the ms-dos call 'getch' which will get a
 * single character from the console
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "fkeys.h"

/*
 * Definitions
 */
#define BUFFER_SIZE 200	/* size of input buffer */
#define MAX_BUFS 5	/* max. number of users */
#define MAX_TASKS 5	/* max. number of tasks */
#define FILENAME_SIZE 50	/* max size of file name */
#define DELAY_FACTOR 500	/* how slow to run tasks */

/* task status flags */
#define TASK_EMPTY  -1	/* task block not in use */
#define TASK_IDLE 0		/* task is not doing anything yet */
#define TASK_GET_INFILE 1 /* task is waiting for input file name */
#define TASK_READING 2	/* task is reading a file */
#define TASK_GET_OUTFILE 3 /* waiting for output file name */
#define TASK_WRITING 4	/* task is writing to file */
#define TASK_STAT 5		/* task is showing status */
#define TASK_KILLED 9	/* task marked to be killed */

/* input buffer structure */
struct InputBuffer {
	char buf[BUFFER_SIZE]; /* actual buffer */
	int bufptr;		 /* location of next position in buffer */
   int stat;		/* status of this user */
};

/* structure for program task */
struct Task {
	int stat;		/* status flag */
   int tty;			/* keyboard running this task */
   char filename[FILENAME_SIZE]; /* file name */
	FILE *fp;		/* file pointer for this task */
   char buf[BUFSIZ]; /* input/output buffer */
   int lineno;		/* current line number */
   int prio;		/* task priority */
   int ctr;			/* count down timer */
};

int get_key(void);
void do_init(void);
void process_key(int ch);
void reg_key(int ch);
void spec_key(int ch);
void do_help(void);
void help_fkeys(void);
void display_status(void);
void display_msg(char *msg);
void process_tasks(void);
int find_free_task(void);
void process_line(void);
void init_task(int loc);
void process_command(char *cmd);
void init_infile(char *fname);
void init_outfile(char *fname);
void process_write(char *outline);
void init_read(void);
void init_write(void);
void do_stop(void);
void do_stat(void);
void execute_read(int id);
void execute_stat(int id);
void display_task_status(int loc);

/*
 * Global Variables
 */

int running = 1;	/* switch indicating we are running */
int cnt=0;			/* times through loop */

struct InputBuffer BufTbl[MAX_BUFS];
int curr_buf = 0;	/* current input buffer */

struct Task TaskTbl[MAX_TASKS];	/* task or process table */


/* logging routine data */
FILE *Logfp = NULL;	/* log file */
char msg[BUFSIZ];	/* common work area */

main()
{
	int ch;		/* keyboard character */

   do_init();	/* initialize things */

   puts("Press F1 for help");

 	while (running)
   { /* loop until done */
   	ch = get_key();
      if (ch > 0)
      {
			process_key(ch);
      }
      process_tasks();
   }

   display_msg("System shut down");
   /* close logfile if needed */
   if (Logfp != NULL) fclose(Logfp);
}

void
do_init(void)
{ /* initialize whatever is necessary */
	int i;

   /* clear the input buffers */
   for (i=0; i < MAX_BUFS; i++)
   {
   	BufTbl[i].bufptr = 0;
      BufTbl[i].buf[0] = '\0';
      BufTbl[i].stat = TASK_IDLE;
   }
   curr_buf = 0;		/* start with buffer zero */

   /* initialize task table */
   for (i=0; i < MAX_TASKS; i++)
   {
   	init_task(i);	/* initialize task entry */
   }

   /* initialize log file */
   Logfp = fopen("event.log", "w");	/* open output file */
   if (Logfp == NULL)
   {
   	puts("Can't open log file.  Logging disabled");
   }
}

void
init_task(int loc)
{ /* initialize task entry at given loc in task table */
	TaskTbl[loc].stat = TASK_EMPTY;	/* not in use */
   TaskTbl[loc].tty = -1;	/* no tty associated with it */
   TaskTbl[loc].filename[0] = '\0';	/* no file name */
   TaskTbl[loc].fp = NULL;	/* no file opened yet */
   TaskTbl[loc].buf[0] = '\0';	/* no data read/written */
   TaskTbl[loc].lineno = -1;	/* no lines read/written */
   TaskTbl[loc].prio = 100;	/* default priority */
   TaskTbl[loc].ctr = 100 * DELAY_FACTOR;		/* count down timer */
}

/*
 * Determine if a key has been pressed and return
 * its code.  If no key, return negative
 */
int
get_key(void)
{
	int c;

   /* if no key is present, return -1 */
   if (!kbhit()) return NOKEY;	/* no key present */

   /*
    * A key was pressed - get scancode
    * If a special key, return the next scancode
    * shiften into the second byte
    */
	c = getch();
	if (c == 0)
	{ /* special key - multiply scan code by X'100' */
		c = getch();	/* special key scancode */
		c <<= 8;			/* shift (multiply) */
	}

	return c; /* return key code */
}

void
process_key(int ch)
{ /* process the pressed key */
   if (isprint(ch))
   { /* regular key */
   	reg_key(ch);
   }
   else
   {
   	spec_key(ch);
   }
}

void
reg_key(int ch)
{ /* insert a character into the buffer */
	/* make sure there is room in buffer */
   if (BufTbl[curr_buf].bufptr > BUFFER_SIZE-1)
   { /* error - buffer full */
   	puts("Buffer full - use backspace or ENTER");
      return;	/* don't add this character */
   }
   BufTbl[curr_buf].buf[BufTbl[curr_buf].bufptr] = ch;
   BufTbl[curr_buf].bufptr++;
}

void
spec_key(int ch)
{
	switch(ch)
   { /* process individual special key */
   case BS: /* backspace - remove previous char */
   	if (BufTbl[curr_buf].bufptr > 0)
      { /* characters in buffer - remove last one */
      	BufTbl[curr_buf].bufptr--;	/* back up one character */
      }
      else puts("The buffer is empty!");
      break;
   case TAB: /* tab - move to next buffer */
   	curr_buf++;		/* increment current buffer */
      /* wrap around to first buffer if at end */
      if (curr_buf >= MAX_BUFS) curr_buf = 0;
      printf("Now using buffer %d\n", curr_buf);
      break;
   case SHIFT_TAB: /* shift tab - move to prev buffer */
      curr_buf--;		/* decrement current buffer */
      /* handle wrap around to last buffer */
      if (curr_buf < 0) curr_buf = MAX_BUFS - 1;
      printf("Now using buffer %d\n", curr_buf);
      break;
   case ENTER: /* end of line */
   	BufTbl[curr_buf].buf[BufTbl[curr_buf].bufptr] = '\0'; /* terminate buffer */
      printf("Buffer %d Line entered >>%s<<\n", curr_buf, BufTbl[curr_buf].buf);
      process_line();		/* execute the command */
      BufTbl[curr_buf].bufptr = 0;		/* reset counter to zero */
      break;
   case ESC: running = 0; break;
   case F1: do_help(); break;
   case F2: help_fkeys(); break;
   case F5:	/* display line so far */
   	BufTbl[curr_buf].buf[BufTbl[curr_buf].bufptr] = '\0';	/* terminate for display */
      printf("Line so far >>%s<<\n", BufTbl[curr_buf].buf);
   	break;
   case F6: display_status(); break;	/* current status */
   case F10: /* reset everything */
   	do_init();
      puts("System reset");
      break;
   default:
		printf("Special key (%d) not supported\n", ch);
   }
}

void
do_help(void)
{ /* display a help message */
	puts("Enter a line of text, then press ENTER.");
   puts("Nothing will be displayed until you press ENTER");
   puts("Press F2 for a list of function keys");
}

void
help_fkeys(void)
{
	puts("F1 - Help");
   puts("F2 - This list");
   puts("F5 - Display line so far");
   puts("F6 - Display Status");
   puts("F10  Reset");
   puts("ENTER - finish line");
   puts("ESC - Exit");
}

void
display_status(void)
{
	int i;

   printf("Current buffer is %d\n", curr_buf);

   printf("Current buffer status:\n");
   for (i=0; i < MAX_BUFS; i++)
   {
   	/* add terminating character to input buffer */
   	BufTbl[i].buf[BufTbl[i].bufptr] = '\0';
   	printf("Buffer %d, stat: %d bufptr=%d >>%s<<\n", i, BufTbl[i].stat,
      		BufTbl[i].bufptr, BufTbl[i].buf);
   }

   /* display task status */
   for (i=0; i < MAX_TASKS; i++)
   { /* only display status of real tasks */
   	if (TaskTbl[i].stat == TASK_EMPTY) continue;	/* not used */
      display_task_status(i);		/* display task status */
   }
}

int
find_free_task(void)
{
	int i;

   for (i=0; i < MAX_TASKS; i++)
   {
   	if (TaskTbl[i].stat == TASK_EMPTY)
      { /* empty slot - initialize it */
      	init_task(i);
         TaskTbl[i].tty = curr_buf;	/* set owner */
         return i;
      }
   }
   /* no free slot found */
   return -1;	/* no free task */
}

void
display_msg(char *str)
{ /* display message and log it to a file */
	long clock;
   struct tm *tm;
   char msg[BUFSIZ];

	clock = time(NULL);	/* get current time */
   tm = localtime(&clock);

   sprintf(msg, "%2.2d:%2.2d.%2.2d %s\n", tm->tm_hour, tm->tm_min,
   					tm->tm_sec, str);
	printf(msg);
   fputs(msg, Logfp);	/* output message */
}

void
process_line(void)
{ /* input line completed - process it */

  	/*
    * based on user's current status,
    * we do something different
    * if writing to a file, this is an output string
    * if waiting for file name, this is the file name
    * for most other states, this is a command
    */
	switch (BufTbl[curr_buf].stat)
   {
   case TASK_IDLE: process_command(BufTbl[curr_buf].buf); break; /* waiting - maybe new command */
   case TASK_GET_INFILE: init_infile(BufTbl[curr_buf].buf); break; /* this is the input file name */
   case TASK_READING: process_command(BufTbl[curr_buf].buf); break; /* reading - maybe stop command */
   case TASK_GET_OUTFILE: init_outfile(BufTbl[curr_buf].buf); break; /* this is the output file name */
   case TASK_WRITING: process_write(BufTbl[curr_buf].buf); break;	/* writing - this is new line */
   default:
   	printf("Internal error - undefined user status (%d)\n",
      		BufTbl[curr_buf].stat);
   }
}

void
process_command(char *cmd)
{ /* user entered a command */
	if (strcmp(cmd, "read") == 0) init_read();
   else if (strcmp(cmd, "write") == 0) init_write();
   else if (strcmp(cmd, "stop") == 0) do_stop();
   else if (strcmp(cmd, "stat") == 0) do_stat();
   else
   {
   	sprintf(msg, "User %d: Unknown command (%s)", curr_buf, cmd);
      display_msg(msg);
   }
}

/*
 * user has issued the read command
 * they should next enter the name of the file to read
 *
 * create a new task for this command,
 * tell the user to enter the file name
 * and wait for their reply
 */
void
init_read(void)
{ /* start a task to read a file */
	int loc;

   /* obtain an empty task block */
   loc = find_free_task();
   if (loc < 0)
   { /* no more tasks - table full */
   	sprintf(msg, "User %d: Can't create new task", curr_buf);
      display_msg(msg);
      return;
   }

   /* task obtained - set it up */
   TaskTbl[loc].stat = TASK_GET_INFILE;	/* wait for input file */
   BufTbl[curr_buf].stat = TASK_GET_INFILE;	/* mark user as waiting */
   TaskTbl[loc].tty = curr_buf;	/* who initiated task */

   /*
    * we are done for now
    * the task will wait until the user enters a line
    * when that happens, we can start reading
    */
    sprintf(msg, "User %d: Read task started (loc %d) waiting for filename",
    					curr_buf, loc);
    display_msg(msg);
    printf("User %d: Enter name of input file\n", curr_buf);
}

/*
 * user has issued write file command
 */
void
init_write(void)
{
	sprintf(msg,"User %d: Writing tasks not supported yet", curr_buf);
	display_msg(msg);
}

/*
 * user has entered the stop command
 *
 * find the task and mark it to be deleted
 */
void
do_stop(void)
{
	int loc;

   /* find what task to stop */
   for (loc=0; loc < MAX_TASKS; loc++)
   { /* find active task owned by this owner */
      if (TaskTbl[loc].stat == TASK_EMPTY) continue; /* task not in use */
   	if (TaskTbl[loc].tty != curr_buf) continue;	/* not same owner */
      if (TaskTbl[loc].stat == TASK_IDLE) continue;	/* not running */
      /* we have found a task to be killed */
      break;
   }

   if (loc < 0 || loc >= MAX_TASKS)
   { /* none found */
   	sprintf(msg, "User %d (do_stop): no active task found", curr_buf);
      display_msg(msg);
      return;
	}

   if (TaskTbl[loc].tty != curr_buf)
   { /* can't stop someone else's task */
   	sprintf(msg, "User %d (do_stop): don't own task id (%d)", curr_buf, loc);
      return;
   }

   /* task ready to be killed - mark it such */
   TaskTbl[loc].stat = TASK_KILLED;

   /*
    * we are donen for now
    * the task will actually get killed during the process tasks
    * loop
    */
    return;
}

/*
 * user entered stat command
 * create task to display the status of the user's tasks
 */
void
do_stat(void)
{
	int loc;

   loc = find_free_task();	/* find free entry in table */
   if (loc < 0)
   {
   	sprintf(msg, "User %d (do_stat): no free task", curr_buf);
      display_msg(msg);
      return;
   }

   /* task created - initialize it */
   TaskTbl[loc].stat = TASK_STAT;
   TaskTbl[loc].tty = curr_buf;
   TaskTbl[loc].lineno = 0;	/* next task to examine */

   /*
    * we are done for now
    * the task program will display task status
    * as it runs
    */
    sprintf(msg, "User %d: stat task created.  task id=%d", curr_buf, loc);
    display_msg(msg);
}

/*
 * The user has entered the name of the input file
 * we can now open the file and begin reading
 *
 * find an outstanding read task that is waiting for a file name
 * fill out task block and mark the task ready to run
 */
void
init_infile(char *fname)
{
	int i;

   for (i=0; i < MAX_TASKS; i++)
   { /* find the task waiting for file name and owned by this user */
   	if (TaskTbl[i].tty != curr_buf) continue;	/* not owned by user */
      if (TaskTbl[i].stat != TASK_GET_INFILE) continue;	/* not waiting */
      /* task found - set it up */
      strcpy(TaskTbl[i].filename, fname);	/* copy file name */
      TaskTbl[i].fp = NULL;		/* file not opened yet */
      TaskTbl[i].stat = TASK_READING;	/* ready to start reading */
      BufTbl[curr_buf].stat = TASK_READING;	/* set user terminal status */
      sprintf(msg, "User %d: Task %d ready for reading file %s", curr_buf, i,
      		TaskTbl[i].filename);
      display_msg(msg);
      return;	/* all done - task will be processed later */
   }
   /* no task found - should not happen */
	sprintf(msg, "User %d (init_read) no task found", curr_buf);
   display_msg(msg);
}

void
init_outfile(char *fname)
{
	display_msg("init_outfile not supported yet");
}

void
process_write(char *outline)
{
	display_msg("process_write not supported");
}

/*
 * This is the routine that "executes" the tasks
 */
void
process_tasks(void)
{
	int i;

   /*
    * loop through task table.  Find active task and execute it
    *
    * Once an active task if found, decrement counter
    * When it turns negative, execute one more statement
    */
   for (i=0; i < MAX_TASKS; i++)
   {
		switch(TaskTbl[i].stat)
      { /* action based on current status */
      case TASK_EMPTY: break;	/* empty - ignore */
      case TASK_IDLE: break;	/* doing nothing - ignore */
      case TASK_GET_INFILE: break; /* waiting for input file name */
      case TASK_READING:
      	TaskTbl[i].ctr--;	/* decrement count down */
         if (TaskTbl[i].ctr < 0)
         { /* ready to run */
            TaskTbl[i].ctr = TaskTbl[i].prio * DELAY_FACTOR;
         	execute_read(i);	/* do another read */
         }
         break;
      case TASK_GET_OUTFILE: break; /* waiting for output file name */
      case TASK_WRITING: break; /* not implemented */
      case TASK_STAT: /* running stat command */
      	TaskTbl[i].ctr--;	/* decrement count down */
         if (TaskTbl[i].ctr < 0)
         { /* ready to run */
         	TaskTbl[i].ctr = TaskTbl[i].prio * DELAY_FACTOR;
            execute_stat(i);	/* execute another statement */
         }
         break;
      case TASK_KILLED: /* terminate this task */
      	TaskTbl[i].stat = TASK_EMPTY;
         sprintf(msg, "User %d: Task %d terminated", curr_buf, i);
         display_msg(msg);
         break;
      default: /* unknown task status */
      	sprintf(msg, "User %d: Unknown task status (%d) for task id %d",
         		curr_buf, TaskTbl[i].stat, i);
         display_msg(msg);
      }
   }
}

/*
 * Execute next statement of the task
 *
 * If file isn't opened yet, open it now
 * read next line into buffer
 */
void
execute_read(int id)
{
	char *p;

	if (TaskTbl[id].fp == NULL)
   { /* file not opened yet */
   	TaskTbl[id].fp = fopen(TaskTbl[id].filename, "r");
      if (TaskTbl[id].fp == NULL)
      { /* open failed */
      	sprintf(msg, "User %d: Can't open file (%s)", curr_buf,
         			TaskTbl[id].filename);
         display_msg(msg);
			TaskTbl[id].stat = TASK_KILLED;	/* kill task */
         return;
      }
   }

   /* read next line from file into buffer */
   TaskTbl[id].lineno++;		/* number of lines read */
   if (fgets(TaskTbl[id].buf, BUFSIZ, TaskTbl[id].fp) == NULL)
   { /* end of file */
   	fclose(TaskTbl[id].fp);	/* close input file */
      TaskTbl[id].fp = NULL;	/* reset file pointer */
      sprintf(msg, "Task %d (user %d): EOF at line %d", id, TaskTbl[id].tty,
      			TaskTbl[id].lineno);
      TaskTbl[id].stat = TASK_KILLED;	/* terminate task */
	}
   else
   { /* line was read */
   	/* strip CR-LF at end of line */
   	for (p=TaskTbl[id].buf; *p != '\0' && *p != '\r' && *p != '\n'; p++);
   	*p = '\0';	/* terminate line */
   	sprintf(msg, "Task %d (user %d): line %d read (%s)", id, TaskTbl[id].tty,
   				TaskTbl[id].lineno, TaskTbl[id].buf);
   }
   display_msg(msg);
}

/*
 * execute next instruction of the stat command
 *
 * id is the location of the stat task
 * lineno is the location of the next task to check
 * display only tasks that are owned by same user
 */
void
execute_stat(int id)
{
	int loc;

   loc = TaskTbl[id].lineno;	/* what task to examine this time */

   /* determine if task is owned by same user */
	if (TaskTbl[loc].tty == TaskTbl[id].tty)
   { /* same owner - display status */
   	display_task_status(loc);
	}

   /*
    * task status checked, set up for next time
    * if all tasks have been checked, terminte the status task
    */
   TaskTbl[id].lineno++;	/* next task to check */
   if (TaskTbl[id].lineno >= MAX_TASKS)
   { /* all done - mark task to be terminated */
   	TaskTbl[id].stat = TASK_KILLED;
   }
}

void
display_task_status(int loc)
{
	sprintf(msg, "Taskid: %d, stat=%d, tty %d, file (%s), line %d, prio %d, ctr %d",
   	loc, TaskTbl[loc].stat, TaskTbl[loc].tty, TaskTbl[loc].filename,
      TaskTbl[loc].lineno, TaskTbl[loc].prio, TaskTbl[loc].ctr);
   display_msg(msg);
}

